use std::collections::HashMap;
use std::env::args_os;
use std::ffi::OsString;

mod spec;

pub use spec::Spec;

pub fn parse(mut spec: Spec) -> ParsedArguments {
    // Split into non-positional and positional arguments.
    let (args, pos_args): (Vec<_>, Vec<_>) = spec
        .args
        .drain(..)
        .partition(|arg| arg.arg_type != spec::ArgType::Pos);

    // Validate positional arguments:
    // - All required positional arguments should precede any optional
    //   positional arguments.
    // - There should be at most a single positional multi-argument, and
    //   it must be at the end.
    {
        let mut met_optional = false;
        let mut met_multi = false;
        for arg in pos_args.iter() {
            if arg.arg_type == spec::ArgType::Pos {
                let is_optional = arg.acceptable_count.0 == 0;
                let is_multi =
                    arg.acceptable_count.1.is_none() || arg.acceptable_count.1.unwrap() > 1;

                if !is_optional && met_optional {
                    panic!("All required positional arguments must precede all optional positional arguments in the argument spec.")
                }
                if met_multi {
                    panic!("There must be at most one positional multi-argument in the argument spec, and it must come last.")
                }

                met_optional |= is_optional;
                met_multi |= is_multi;
            }
        }
    }

    // Parse!
    // TODO: optimize by first creating a hash map from flag strings to
    // argument indices.  Right now this is an `O(NM)` algorithm, with N
    // being the number of arguments in the spec and M being the number
    // of arguments passed by the user.  We can even further optimize it
    // by first checking against the maximum length of our long arguments,
    // so we don't end up hashing really long user arguments
    // unnecessarily for the check.
    let mut pos_i = 0; // Index of the positional argument we're at.
    let mut pos_i_count = 0; // Number of positional arguments we've parsed at the current positional argument index.
    let mut parsed = ParsedArguments {
        arguments: Vec::new(),
        id_map: HashMap::new(),
    };
    let mut args_in = args_os();
    let _ = args_in.next(); // Skip the first argument, which is the call to the executable.

    'outer: while let Some(arg_in) = args_in.next() {
        // Check for flags and non-positional arguments.
        if let Some(arg_in_str) = arg_in.to_str() {
            if arg_in_str.starts_with("--") {
                // Long.
                for arg in args.iter() {
                    for long_flag in arg.long_flags.iter() {
                        if arg_in_str == long_flag {
                            match arg.arg_type {
                                spec::ArgType::Flag => parsed.push_arg(arg.id.clone(), None),

                                spec::ArgType::Arg => {
                                    if let Some(value) = args_in.next() {
                                        parsed.push_arg(arg.id.clone(), Some(value));
                                    } else {
                                        todo!("Handle error: expected value after argument flag.");
                                    }
                                }

                                spec::ArgType::Pos => unreachable!(),
                            }
                            continue 'outer;
                        }
                    }
                }
                todo!("Handle error: no long argument matched the passed argument.");
            } else if arg_in_str.starts_with("-") {
                // Short.
                let mut remainder = &arg_in_str[1..];

                // First check arguments that take values.
                for arg in args.iter().filter(|a| a.arg_type == spec::ArgType::Arg) {
                    for short_flag in arg.short_flags.iter() {
                        if remainder == short_flag {
                            if let Some(value) = args_in.next() {
                                parsed.push_arg(arg.id.clone(), Some(value));
                            } else {
                                todo!("Handle error: expected value after argument flag.");
                            }
                            continue 'outer;
                        }
                    }
                }

                // Then check boolean flags.  There can be multiple
                // present, so we progressively chop off the front as we
                // find matches until nothing remains.
                'restart_args: while !remainder.is_empty() {
                    for arg in args.iter().filter(|a| a.arg_type == spec::ArgType::Flag) {
                        for short_flag in arg.short_flags.iter() {
                            if remainder.starts_with(short_flag) {
                                remainder = &remainder[short_flag.len()..];
                                parsed.push_arg(arg.id.clone(), None);
                                continue 'restart_args;
                            }
                        }
                    }
                    todo!(
                        "Handle error: no short argument matches the next flag in \"{}\".",
                        remainder
                    );
                }
                continue 'outer;
            }
        }

        if pos_i < pos_args.len() {
            let arg = &pos_args[pos_i];
            parsed.push_arg(arg.id.clone(), Some(arg_in));
            pos_i_count += 1;

            if let Some(max_count) = arg.acceptable_count.1 {
                if pos_i_count == max_count {
                    pos_i += 1;
                    pos_i_count = 0;
                }
            }
        } else {
            todo!("Handle error: too many positional arguments.");
        }
    }

    if pos_i < pos_args.len() && pos_i_count < pos_args[pos_i].acceptable_count.0 {
        todo!("Handle error: not enough positional arguments.");
    }

    parsed
}

/// Parsed command line arguments.
#[derive(Debug, Clone)]
pub struct ParsedArguments {
    // A list of all passed arguments, in the order they were
    // passed.  Each item is a `(argument_id, argument_value)` tuple.
    // Boolean flags do not have a value.
    arguments: Vec<(String, Option<OsString>)>,

    // Maps from an argument's ID to the indices in `arguments` where it
    // is stored.
    id_map: HashMap<String, Vec<usize>>, // Argument ID -> index list
}

impl ParsedArguments {
    fn push_arg(&mut self, id: String, value: Option<OsString>) {
        assert!(!id.is_empty());

        if !self.id_map.contains_key(&id) {
            self.id_map.insert(id.clone(), Vec::new());
        }
        self.id_map.get_mut(&id).unwrap().push(self.arguments.len());

        self.arguments.push((id, value));
    }
}
