use std::collections::HashSet;

/// Command line argument specification.
#[derive(Debug, Clone)]
pub struct Spec {
    pub(crate) name: String,    // Application name.
    pub(crate) version: String, // Application version.
    pub(crate) args: Vec<Arg>,

    // Used to ensure we don't get duplicate arguments.
    id_set: HashSet<String>,
    long_set: HashSet<String>,
    short_set: HashSet<String>,
}

impl Spec {
    /// Create a new argument specification.
    ///
    /// `name` and `version` are the name and version of the software,
    /// respectively.
    #[must_use]
    pub fn new(name: String, version: String) -> Spec {
        Spec {
            name: name,
            version: version,

            args: Vec::new(),

            id_set: HashSet::new(),
            long_set: HashSet::new(),
            short_set: HashSet::new(),
        }
    }

    /// Add a flag (bool) argument.
    ///
    /// - `id`: the argument identifier, used for fetching argument
    ///   matches.
    /// - `flags`: the long and/or short argument flag strings.  Must be
    ///   in the form "-f" or "--flag".  You can pass as many as you
    ///   like, all of which will be considered equivalent during
    ///   parsing.  But there must be at least one.
    /// - `doc`: the documentation string to use in the generated help.
    ///   Pass an empty string to indicate no documentation.
    #[must_use]
    pub fn add_flag(mut self, id: &str, flags: &[&str], doc: &str) -> Self {
        let (long_flags, short_flags) = self.validate_and_process_arg(id, flags);

        self.args.push(Arg {
            arg_type: ArgType::Flag,
            id: id.into(),
            value_label: String::new(),
            long_flags: long_flags,
            short_flags: short_flags,
            acceptable_count: (0, None),
            doc: doc.into(),
        });

        self
    }

    /// Add a standard argument, that takes a value.
    #[must_use]
    pub fn add_argument(
        mut self,
        id: &str,
        flags: &[&str],
        doc: &str,
        value_label: &str,
        required: bool,
    ) -> Self {
        let (long_flags, short_flags) = self.validate_and_process_arg(id, flags);

        self.args.push(Arg {
            arg_type: ArgType::Arg,
            id: id.into(),
            value_label: value_label.into(),
            long_flags: long_flags,
            short_flags: short_flags,
            acceptable_count: (if required { 1 } else { 0 }, None),
            doc: doc.into(),
        });

        self
    }

    /// Add a positional argument.
    ///
    /// Unlike flags and standard arguments, positional arguments are
    /// parsed in the order they're added.  Because of their nature,
    /// they have some additional considerations:
    ///
    /// - All required positional arguments must precede all optional
    ///   positional arguments.
    /// - There can at most be a single positional multi-argument,
    ///   which must come last.  (See `add_positional_multi_argument()`.)
    #[must_use]
    pub fn add_positional_argument(
        mut self,
        id: &str,
        doc: &str,
        value_label: &str,
        required: bool,
    ) -> Self {
        let (_, _) = self.validate_and_process_arg(id, &[]);

        self.args.push(Arg {
            arg_type: ArgType::Pos,
            id: id.into(),
            value_label: value_label.into(),
            long_flags: Vec::new(),
            short_flags: Vec::new(),
            acceptable_count: (if required { 1 } else { 0 }, Some(1)),
            doc: doc.into(),
        });

        self
    }

    #[must_use]
    pub fn add_positional_multi_argument(
        mut self,
        id: &str,
        doc: &str,
        value_label: &str,
        required: bool,
    ) -> Self {
        let (_, _) = self.validate_and_process_arg(id, &[]);

        self.args.push(Arg {
            arg_type: ArgType::Pos,
            id: id.into(),
            value_label: value_label.into(),
            long_flags: Vec::new(),
            short_flags: Vec::new(),
            acceptable_count: (if required { 1 } else { 0 }, None),
            doc: doc.into(),
        });

        self
    }

    //----------------

    /// Returns (long, short) pair, each of which is a Vec of argument strings with
    /// the leading hyphens stripped off.
    fn validate_and_process_arg(&mut self, id: &str, flags: &[&str]) -> (Vec<String>, Vec<String>) {
        if self.id_set.contains(id) {
            panic!(
                "Error: attempted to add argument with a duplicate ID \"{}\".",
                id
            );
        }
        self.id_set.insert(id.into());

        let mut long_flags = Vec::new();
        let mut short_flags = Vec::new();

        for &flag in flags {
            // Ensure no whitespace.
            if flag.len() != flag.trim().len() || flag.split_whitespace().count() > 1 {
                panic!(
                    "Error: attempted to add argument \"{}\" which contains whitespace.",
                    flag
                );
            }
            // Long flags.
            else if flag.starts_with("--") && flag.len() > 2 {
                if self.long_set.contains(flag) {
                    panic!(
                        "Error: attempted to add duplicate long argument \"{}\".",
                        flag
                    );
                }
                self.long_set.insert(flag.into());
                long_flags.push(flag.into());
            }
            // Check if it's a valid short flag.
            // Note: in theory we should be checking to verify that
            // the flag is only one character long.  But because of
            // graphemes that's complicated, and it's not really
            // worth all the code.  So instead we just rely on client
            // code doing the right thing.
            else if flag.starts_with("-") && flag.len() > 1 {
                if self.short_set.contains(flag) {
                    panic!(
                        "Error: attempted to add duplicate short argument \"{}\".",
                        flag
                    );
                }
                self.short_set.insert(flag.into());
                short_flags.push((&flag[1..]).into());
            }
            // Not a valid flag.
            else {
                panic!(
                    "Error: attempted to add argument \"{}\", which isn't a valid argument string.",
                    flag
                )
            }
        }

        (long_flags, short_flags)
    }
}

//-------------------------------------------------------------

#[derive(Debug, Copy, Clone, Eq, PartialEq)]
pub(crate) enum ArgType {
    Flag, // Boolean flag (present or absent)
    Arg,  // Standard flag+value argument, like `-i input_file`.
    Pos,  // Positional argument.
}

/// Argument specification.
#[derive(Debug, Clone)]
pub(crate) struct Arg {
    pub(crate) arg_type: ArgType,
    pub(crate) id: String,
    pub(crate) value_label: String,

    // Long and short versions of the argument flag.  E.g. "--curve" and
    // "-c", but without the leading dashes.
    pub(crate) long_flags: Vec<String>,
    pub(crate) short_flags: Vec<String>,

    // How many instances of the argument can be present, specified
    // as a range.
    //
    // For example:
    // - (0, None): An argument that can show up any number of times,
    //              including not at all.
    // - (0, 1):    An argument that can either be absent or show up
    //              precisely once.
    // - (1, 1):    An argument that must show up precisely once.
    // - (1, None): An argument that must show up at least once.
    // - (2, 9):    An argument that must show up at least twice, but
    //              no more than 9 times.
    pub(crate) acceptable_count: (usize, Option<usize>),

    // Documentation string, for generated help.
    pub(crate) doc: String,
}
